/* $Id: common.c,v 1.14 2002/11/08 15:31:51 daniel Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *    common: Functions used by other parts of Repligard
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "midgard/midgard.h"
#include "read.h"
#include "rpconfig.h"

#ifndef S_ISREG
#define S_ISREG(mode)	(((mode)&S_IFMT) == S_IFREG)
#endif

#ifdef WIN32
typedef unsigned short mode_t;
#endif

#ifndef WIN32
extern uid_t realuserid, suiduserid;
#endif

void repligard_build_guid_hash_add(midgard * mgd,GHashTable *guid_hash, const gchar *guid )
{ 
	 
	midgard_res *res;
	WGuidCacheEntry *entry;

	// Gather GUIDs from the database for better performance 
 
	g_assert(guid_hash != NULL);
        if (configuration.verbose > 1)
    	         	  	 g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
			   	 "calling SELECT");
	res =
	   mgd_ungrouped_select(mgd, "id,guid,realm,changed,updated,action",
				"repligard", "guid=$q", NULL, guid);

	while (res && mgd_fetch(res)) {
	
		entry = g_new(WGuidCacheEntry, 1);
		g_assert(entry != NULL);
		entry->changed = g_strdup(mgd_column(res, "changed"));
		entry->updated = g_strdup(mgd_column(res, "updated"));
		entry->guid = g_strdup(mgd_column(res, "guid"));
		if (configuration.verbose > 2)
    	         	  	 g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
			   	 "got %s %s %s", entry->guid, entry->updated,entry->changed);
		entry->deleted = FALSE;
		if (g_strcasecmp("delete", mgd_column(res, "action")) == 0)
			entry->deleted = TRUE;
		entry->replicated = FALSE;
		g_hash_table_insert(guid_hash,
				    g_strconcat(mgd_column(res, "realm"), " ",
						mgd_column(res, "id"), NULL),
				    entry);
	}
	if (res)
		mgd_release(res);
  
}


 
GHashTable *repligard_build_guid_hash(midgard * mgd, RState state)
{
	GHashTable *guid_hash;
	midgard_res *res;
	WGuidCacheEntry *entry;

	// Gather GUIDs from the database for better performance 
	guid_hash = g_hash_table_new(rpd_honeyman_hash, g_str_equal);

	g_assert(guid_hash != NULL);
       if (state == R_IMPORT)     return guid_hash;
	res =
	   mgd_ungrouped_select(mgd, "id,guid,realm,changed,updated,action",
				"repligard", NULL, NULL);

	while (res && mgd_fetch(res)) {
		entry = g_new(WGuidCacheEntry, 1);
		g_assert(entry != NULL);
		entry->changed = g_strdup(mgd_column(res, "changed"));
		entry->updated = g_strdup(mgd_column(res, "updated"));
		entry->guid = g_strdup(mgd_column(res, "guid"));
		entry->deleted = FALSE;
		if (g_strcasecmp("delete", mgd_column(res, "action")) == 0)
			entry->deleted = TRUE;
		entry->replicated = FALSE;
		g_hash_table_insert(guid_hash,
				    g_strconcat(mgd_column(res, "realm"), " ",
						mgd_column(res, "id"), NULL),
				    entry);
	}
	if (res)
		mgd_release(res);
	return guid_hash;
}

/* BASE64 encoding support for BLOBs */

/* BASE64 algorithm implementation is based on code written by 

			    by John Walker
		       http://www.fourmilab.ch/

		This program is in the public domain.
*/
static FILE *infile, *outfile;
static gzFile *ofile;
static GByteArray *instream;
static guint8 iobuf[256];	/* I/O buffer */
static int iolen = 0;		/* Bytes left in I/O buffer */
static int iocp = 256;		/* Character removal pointer */
static int iodeccp = 0;
static int ateof = FALSE;	/* EOF encountered */
static guint8 dtable[256];	/* Encode / decode table */
static int linelength = 0;	/* Length of encoded output line */
static int errcheck = TRUE;	/* Check decode input for errors ? */
#define LINELEN 64
/*  INBUF  --  Fill input buffer with data  */

static int inbuf(void)
{
	int l;

	if (ateof) {
		return FALSE;
	}
	l = fread(iobuf, 1, 256, infile);	/* Read input buffer */
	if (l <= 0) {
		if (ferror(infile)) {
			g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
			      "Cannot open BLOB file");
			myexit(1);
		}
		ateof = TRUE;
		return FALSE;
	}
	iolen = l;
	iocp = 0;
	return TRUE;
}

/*  INCHAR  --	Return next character from input  */

static int inchar(gboolean from)
{
	if (from == TRUE) {
		if (iocp >= iolen) {
			if (!inbuf()) {
				return EOF;
			}
		}
		return iobuf[iocp++];
	}
	else {
		if (iodeccp >= instream->len)
			return EOF;
		return (int) instream->data[iodeccp++];
	}
}

/*  OCHAR  --  Output an encoded character, inserting line breaks
	       where required.	*/

static void ochar(int c)
{
	if (linelength >= LINELEN) {
		gzputs(ofile, "\n");
		linelength = 0;
	}
	gzputc(ofile, c);
	linelength++;
}

/*  ENCODE  --	Encode binary file into base64.  */

static void encode(void)
{
	int i, hiteof = FALSE;

	/*  Fill dtable with character encodings.  */

	for (i = 0; i < 26; i++) {
		dtable[i] = 'A' + i;
		dtable[26 + i] = 'a' + i;
	}
	for (i = 0; i < 10; i++) {
		dtable[52 + i] = '0' + i;
	}
	dtable[62] = '+';
	dtable[63] = '/';

	while (!hiteof) {
		guint8 igroup[3], ogroup[4];
		int c, n;

		igroup[0] = igroup[1] = igroup[2] = 0;
		for (n = 0; n < 3; n++) {
			c = inchar(TRUE);
			if (c == EOF) {
				hiteof = TRUE;
				break;
			}
			igroup[n] = (guint8) c;
		}
		if (n > 0) {
			ogroup[0] = dtable[igroup[0] >> 2];
			ogroup[1] =
			   dtable[((igroup[0] & 3) << 4) | (igroup[1] >> 4)];
			ogroup[2] =
			   dtable[((igroup[1] & 0xF) << 2) | (igroup[2] >> 6)];
			ogroup[3] = dtable[igroup[2] & 0x3F];

			/* Replace characters in output stream with "=" pad
			   characters if fewer than three characters were
			   read from the end of the input stream. */

			if (n < 3) {
				ogroup[3] = '=';
				if (n < 2) {
					ogroup[2] = '=';
				}
			}
			for (i = 0; i < 4; i++) {
				ochar(ogroup[i]);
			}
		}
	}
	gzputs(ofile, "\n");
}

/*  INSIG  --  Return next significant input  */

static int insig(void)
{
	int c;

	 /*CONSTANTCONDITION*/ while (TRUE) {
		c = inchar(FALSE);
		if (c == EOF || (c > ' ')) {
			return c;
		}
	}
 /*NOTREACHED*/}

/*  DECODE  --	Decode base64.	*/

static void decode(void)
{
	int i;

	for (i = 0; i < 255; i++) {
		dtable[i] = 0x80;
	}
	for (i = 'A'; i <= 'Z'; i++) {
		dtable[i] = 0 + (i - 'A');
	}
	for (i = 'a'; i <= 'z'; i++) {
		dtable[i] = 26 + (i - 'a');
	}
	for (i = '0'; i <= '9'; i++) {
		dtable[i] = 52 + (i - '0');
	}
	dtable['+'] = 62;
	dtable['/'] = 63;
	dtable['='] = 0;

	 /*CONSTANTCONDITION*/ while (TRUE) {
		guint8 a[4], b[4], o[3];

		for (i = 0; i < 4; i++) {
			int c = insig();

			if (c == EOF) {
				if (errcheck && (i > 0)) {
					g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
					      "Input stream for BLOB incomplete.");
					myexit(1);
				}
				return;
			}
			if (dtable[c] & 0x80) {
				if (errcheck) {
					g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
					      "Illegal character '%c' in input file.",
					      c);
					myexit(1);
				}
				/* Ignoring errors: discard invalid character. */
				i--;
				continue;
			}
			a[i] = (guint8) c;
			b[i] = (guint8) dtable[c];
		}
		o[0] = (b[0] << 2) | (b[1] >> 4);
		o[1] = (b[1] << 4) | (b[2] >> 2);
		o[2] = (b[2] << 6) | b[3];
		i = a[2] == '=' ? 1 : (a[3] == '=' ? 2 : 3);
		fwrite(o, 1, i, outfile);
		if (i < 3) {
			return;
		}
	}
}

void repligard_decode_blob(GByteArray * blob, const gchar * filename)
{
	gchar *fname;
#ifndef WIN32
	mode_t oldmask;
#endif
	if ((filename == NULL) || (blob == NULL))
		return;
	if (*filename == '\0')
		return;
	fname = g_strdup_printf("%s/%s", configuration.blobdir, filename);
#ifndef WIN32
	if (realuserid != 0) setuid(suiduserid);
	oldmask = umask(02);
#endif
	outfile = fopen(fname, "wb");
	if (!outfile) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		      "Error writing BLOB: cannot open file %s for writing",
		      fname);
		g_free(fname);
		myexit(1);
	}
	instream = blob;
	iodeccp = 0;
	decode();
	fclose(outfile);
#ifndef WIN32
	if (realuserid != 0) setuid(realuserid);
#endif
	g_free(fname);
#ifndef WIN32
	(void) umask(oldmask);
#endif
}

void repligard_encode_blob(const gchar * filename, gzFile * outfile)
{
	gchar *fname;
	struct stat filestat;
	if ((filename == NULL) || (outfile == NULL))
		return;
	if (*filename == '\0')
		return;
	ofile = outfile;
	fname = g_strdup_printf("%s/%s", configuration.blobdir, filename);
	
#ifndef WIN32
	if (realuserid != 0) setuid(suiduserid);
#endif
	if (stat(fname, &filestat) == -1) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		      "Cannot open BLOB container '%s' at '%s'.",
		      filename, configuration.blobdir);
			myexit(1);
	}
	if (!S_ISREG(filestat.st_mode)) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		      "Cannot open BLOB container '%s' because "
		      "it is not a regular file in '%s'.",
		      filename, configuration.blobdir);
			myexit(1);
	}
	if ((infile = fopen(fname, "r")) == NULL) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		      "Cannot open BLOB container '%s' at '%s'.",
		      filename, configuration.blobdir);
			myexit(1);
	}
	iocp = 0;
	iolen = 0;
	ateof = FALSE;
	linelength = 0;
	encode();
	fclose(infile);
#ifndef WIN32
	if (realuserid != 0) setuid(realuserid);
#endif
	
	g_free(fname);
}
