/* $Id: common.h,v 1.9 2002/09/12 16:28:46 daniel Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *    common: common type definitions for import and export
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _REPLIGARD_COMMON_H_
#define _REPLIGARD_COMMON_H_

#include <zlib.h>
#include "midgard/midgard.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>

#ifdef __cplusplus
extern "C"
{
#endif				/* __cplusplus */
typedef enum
{
  R_NONE,
  R_IMPORT,
  R_EXPORT
}
RState;


typedef struct
{
  gzFile *fd;
  GHashTable *guid_hash;
  GHashTable *schema;
  RSchemaCfg *schemacfg;
  midgard *mgd;
  GList *dependencies;
  gboolean expandtree;
}
WExport;

typedef struct
{
  GHashTable *guid_hash;
  GHashTable *schema;
  RSchemaCfg *schemacfg;
  midgard *mgd;
}
WImport;


typedef struct
{
  gzFile *fd;
  midgard_res *res;
  GHashTable *guid_hash;
  WExport *exp_data;
}
WObjectFD;

typedef struct
{
  midgard_res *res;
  GHashTable *otype;
  gchar *guid;
  const gchar *ntype;
}
WObjectType;

typedef struct
{
  gchar *changed;
  gchar *updated;
  gchar *guid;
  gboolean deleted;
  gboolean replicated;
}
WGuidCacheEntry;

typedef struct
{
  gchar *name;
  gchar *guid;
  gchar *changed;
  GHashTable *object;
}
RDelayEntry;

void repligard_import_resource (WImport * imp_data,
				GHashTable * object,
				const gchar * object_name, const gchar * guid,
				const gchar * changed);

void repligard_delete_resource (WImport * imp_data,
				GHashTable * object,
				const gchar * object_name, const gchar * guid,
				const gchar * deleted);

int repligard_create_fake_resource (WImport * imp_data,
				    gchar * object_name, gchar * guid);

GHashTable *repligard_build_guid_hash (midgard * mgd, RState state);
void repligard_build_guid_hash_add(midgard * mgd,GHashTable *guid_hash, const
gchar *guid );

  /* Decodes BASE64-encoded blob string into byte array */
void repligard_decode_blob (GByteArray * blob, const gchar *filename);

  /* Encodes blob using BASE64 encoding and writes to output stream */
  /* Filename should be relative to BLOBs directory */
void repligard_encode_blob (const gchar * filename, gzFile * outfile);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _REPLIGARD_COMMON_H_ */
