/* $Id: exporthook.c,v 1.3 2001/07/05 21:05:04 ab Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 * exporthook: abstract API for export hooks
 *
 * Copyright (C) 2001 The Midgard Project Ry
 * Copyright (C) 2001 Alexander Bokovoy 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "exporthook.h"
#include "rpconfig.h"

  /* Set export hook to the specified one */
  /* Repligard supports only one global export hook. It means that only one language
     can be used for scripting checking procedures. It isn't that hard to add per object type
     langauge support but usually it is inconvinient from both performance and 
     user experience sides. Almost all things could be done using one specific language.
   */
void repligard_set_export_hook (RExportHook hook) {
  if(hook) {
    configuration.export_hook = hook;
  }
}

  /* Clear export hook */
void repligard_clear_export_hook (void) {
  configuration.export_hook = NULL;
}

  /* Run export hook */
gint repligard_run_export_hook (WObjectType *object_type) {
  const gchar *script;
  if((object_type != NULL) && 
     (configuration.export_hook != NULL) && (configuration.export_types != NULL)) {
    /* Get script associated with this object type */
    script = g_hash_table_lookup(configuration.export_types, object_type->ntype);
    /* Run script. Hook is responsible to return 0 if script is NULL */
    /* which means that object must be replicated */
    return configuration.export_hook(object_type, script);
  }
  /* Return 0 if hook isn't set. This is absolutely normal situation. */
  return 0;
}

void repligard_add_export_type (const gchar *name, const gchar *script) {
  gchar *old_script, *old_name;
  gboolean flag;
  if(configuration.export_types == NULL) {
    configuration.export_types = g_hash_table_new(rpd_honeyman_hash, g_str_equal);
  }
  flag = g_hash_table_lookup_extended(configuration.export_types, name, (gpointer*) &old_name, (gpointer*) &old_script);
  if(old_script != NULL) {
    g_hash_table_remove(configuration.export_types, name);
    g_free(old_name); g_free(old_script);
  }
  /* NULL script means 'delete hook for this type' */
  if(script != NULL)
    g_hash_table_insert(configuration.export_types, g_strdup(name), (gchar*) script);
}

static gboolean
rpd_remove_type_foreach (gpointer key, gpointer value, gpointer user_data) {
  if(value) {
    g_free(value);
  }
  g_free(key);

  return TRUE;
}

void repligard_destroy_export_types (void) {
  if(configuration.export_types != NULL) {
    g_hash_table_foreach_remove(configuration.export_types, rpd_remove_type_foreach, NULL);
    g_hash_table_destroy(configuration.export_types);
    configuration.export_types = NULL;
  }
}
