/* $Id: exporthook.h,v 1.3 2002/09/12 16:29:57 daniel Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 * exporthook: abstract API for export hooks
 *
 * Copyright (C) 2001 The Midgard Project Ry
 * Copyright (C) 2001 Alexander Bokovoy 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _REPLIGARD_EXPORTHOOK_H_
#define _REPLIGARD_EXPORTHOOK_H_

#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>
#include "midgard/midgard.h"
#include "schema.h"
#include "common.h"

#ifdef __cplusplus
extern "C"
{
#endif				/* __cplusplus */

  /* Functional for evaluating an object in context of the script */
  /* The functional should return script result as integer value */
  /* The meaning of the result is the same as for conventional */
  /* programs, i.e. 0 means success, 1 -- an error. */
  /* As result, object is exported then and only then when functional returns 0 */
  typedef gint (*RExportHook) (WObjectType *object_type, const gchar *script);

  /* Set export hook to the specified one */
  /* Repligard supports only one global export hook. It means that only one language
     can be used for scripting checking procedures. It isn't that hard to add per object type
     langauge support but usually it is inconvinient from both performance and 
     user experience sides. Almost all things could be done using one specific language.
   */
  void repligard_set_export_hook (RExportHook hook);

  /* Clear export hook */
  void repligard_clear_export_hook (void);

  /* Run export hook */
  gint repligard_run_export_hook (WObjectType *object_type);
  
  /* Add export script for the specified object type 'name' */
  /* This function creates export types' hash if needed */
  void repligard_add_export_type (const gchar *name, const gchar *script);

  /* Destroy export types' hash */
  void repligard_destroy_export_types (void);

#ifdef __cplusplus
}
#endif				/* __cplusplus */

#endif				/* _REPLIGARD_EXPORTHOOK_H_ */
