/* $Id: rpconfig.h,v 1.12 2002/09/12 16:29:57 daniel Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *  rpconfig: read Repligard configuration file
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _REPLIGARD_RPCONFIG_H_
#define _REPLIGARD_RPCONFIG_H_

#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>
#include "midgard/midgard.h"
#include "schema.h"
#include "exporthook.h"

#ifdef __cplusplus
extern "C"
{
#endif				/* __cplusplus */

#define R_TAG_RCFG_REPLIGARD "Repligard"
#define R_TAG_RCFG_DATABASE  "database"
#define R_TAG_RCFG_LOGIN     "login"
#define R_TAG_RCFG_REPLICATE "replicate"
#define R_TAG_RCFG_RESOURCE  "resource"
#define R_TAG_RCFG_INCLUDE   "include"
#define R_TAG_RCFG_EXPORTHOOK "exporthook"

#define R_ATT_RCFG_NAME      "name"
#define R_ATT_RCFG_USERNAME  "username"
#define R_ATT_RCFG_PASSWORD  "password"
#define R_ATT_RCFG_ENCODING  "encoding"
#define R_ATT_RCFG_BLOBDIR   "blobdir"
#define R_ATT_RCFG_SCHEMA    "schema"
#define R_ATT_RCFG_ID        "id"
#define R_ATT_RCFG_TYPE      "type"
#define R_ATT_RCFG_GUID      "guid"
#define R_ATT_RCFG_ALL       "all"

  typedef struct
  {
    gchar *type;
    gchar *id;
    gboolean guid;
  }
  RListEntry;

  typedef struct
  {
    /* Path to schema */
    gchar *schema;
    /* Database information */
    gchar *dbname, *dbusername, *dbpassword, *dbencoding;

    /* Login information */
    gchar *username, *password;

    /* Root directory for BLOBs */
    gchar *blobdir;

    /* Replication information */
    gboolean all;
    /* 
       This list keeps information about resources which 
       should be replicated. It includes either resource type:local_id
       or GUID of resource.
       If there are no <resource> sections in the <replicate>, then
       this list will be NULL. If 'all' attribute is set
       for <replicate>, then all resources will be replicated.
     */
    GList *resource;
    /* Do diff against last updated or just dump all records in 'replicate all' mode? */
    gboolean diff;
    gint verbose;
    gboolean sitegroup;
    gboolean nopasswords;
    gboolean quote_as_cdata;
    gchar *fromdate;
    
    /* Export hook: all data will be passed through it before export if set */
    RExportHook export_hook;
    
    /* Export scripts hashed by object types. 
       This table is ruled by <exporthook type="foo"/> directive.
       You can set one hook per object type. It is usually flexible enough.
    */
    GHashTable *export_types;
    
  }
  RConfiguration;

  extern RConfiguration configuration;
  void repligard_init (void);
  void repligard_done (void);
  void repligard_read_config (const gchar * filename);
  /* Helper for list management */
  void rpd_remove_stringlist (gpointer value, gpointer user_data);
  void myexit(int errcode);

#ifdef __cplusplus
}
#endif				/* __cplusplus */

#endif				/* _REPLIGARD_RPCONFIG_H_ */
