/* $Id: schema.h,v 1.5 2002/09/12 16:29:57 daniel Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *    schema: utility functions for parsing Repligard config file
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Aurora - Alexander Bokovoy
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _REPLIGARD_SCHEMA_H_
#define _REPLIGARD_SCHEMA_H_

#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>

#ifdef __cplusplus
extern "C"
{
#endif				/* __cplusplus */

#define R_SCHEMA_NS "http://www.midgard-project.org/repligard/1.4"
#define R_TAG_SCHEMA_SCHEMA "Schema"
#define R_TAG_SCHEMA_TYPE   "type"
#define R_TAG_SCHEMA_FIELD  "field"
#define R_TAG_SCHEMA_LINK   "link"
#define R_TAG_SCHEMA_RLINK   "rlink"

#define R_ATT_SCHEMA_NAME   "name"
#define R_ATT_SCHEMA_LINK   "link"
#define R_ATT_SCHEMA_TYPE   "type"
#define R_ATT_SCHEMA_REVERSE "reverse"
 
#define R_ATT_BLOB_LOCID    "locid"
#define R_ATT_SCHEMA_REVERSE_VAL "no"

  typedef enum
  {
    R_STRING_FIELD = 1,
    R_TEXT_FIELD,
    R_INTEGER_FIELD,
    R_PASSWORD_FIELD,
    R_LINK_FIELD,
    R_LINK_FIELD_NODEPS,
    R_RLINK_FIELD,
    R_BLOB_FIELD
  }
  RSchemaFieldType;

  typedef struct
  {
    RSchemaFieldType type;	/* Type of the field */
    gchar *link;		/* Link value (if type is R_LINK_FIELD or R_RLINK_FIELD) */
  }
  RSchemaField;

  typedef enum
  {
    R_FLAT_OBJECT = 1,		/* Object of this type doesn't have links from the objects of the same type */
    R_TREE_OBJECT,		/* Object of this type has links from the objects of the same type */
    R_ACTIVE_OBJECT,		/* Object of this type has link to an object which nature is determined at run time */
    R_FILE_OBJECT		/* Object should be treated as a link to the external resource like BLOB */
  }
  RSchemaObjectType;

  typedef struct
  {
    RSchemaObjectType type;
    GHashTable *object;
  }
  RSchemaType;

  typedef struct
  {
    GHashTable *schema;		/* Contains entries of RSchemaType */
    GHashTable *links;
  }
  RSchemaCfg;

  /* Hash function for Repligard */
  guint rpd_honeyman_hash (gconstpointer key);

  /* Initializes Repligard type schema using external definition file in XML format */
  RSchemaCfg *repligard_init_schema (const gchar * filename);

  /* Frees previously initialized Repligard type schema */
  void repligard_done_schema (RSchemaCfg * schema);

  /* Returns field type based on its literal representation */
  RSchemaFieldType repligard_field_from_string (const gchar * str);

  /* Returns literal representation of field type */
  const gchar *repligard_string_from_field (RSchemaFieldType t);

  /* Returns field description for the object directly from Repligard schema */
  RSchemaField *repligard_get_qualified_field (GHashTable * schema,
					       const gchar * object,
					       const gchar * field);

  /* Returns hash describing object as a series of fields */
  GHashTable *repligard_get_object (GHashTable * schema,
				    const gchar * object);

  /* Returns a record describing the object */
  RSchemaType *repligard_get_object_type (GHashTable * schema,
					  const gchar * object);

  /* Returns field description for the object */
  RSchemaField *repligard_get_field (GHashTable * object,
				     const gchar * field);

  /* Get attribute value from the attribute list */
  const gchar *repligard_get_attribute (const char *name, const char **atts);

#ifdef __cplusplus
}
#endif				/* __cplusplus */

#endif				/* _REPLIGARD_SCHEMA_H_ */
