/* $Id: filetemplate.c,v 1.10 2002/09/12 16:34:43 daniel Exp $
 *
 * midgard-lib: database access for Midgard clients
 *
 * Copyright (C) 2000 The Midgard Project ry
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "midgard/midgard.h"

#if HAVE_MIDGARD_FILETEMPLATES

#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <stdlib.h>
#include <string.h>
#ifdef WIN32
#include "win32/win95nt.h"
#else
#include <sys/param.h>
#endif
#include <errno.h>
#include <stdio.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>

#define MGD_FT_ELTNAME_PAGE         "page"
#define MGD_FT_ELTNAME_PAGE_NAME    "title"
#define MGD_FT_ELTNAME_PAGE_CONTENT "content"
#define MGD_FT_ELTNAME_ELT          "element"
#define MGD_FT_ELTNAME_ELT_NAME     "name"
#define MGD_FT_ELTNAME_STOP         "stop"
#define MGD_FT_ELTNAME_SHS          "skiphoststyle"
#define MGD_FT_ELTNAME_AUTH         "auth"

char *mgd_strrchr(char *str, char c, int *p)
{
	char *pos = str + *p;

	if (*p < 0)
		return NULL;
	if (!str)
		return NULL;

	while (pos >= str && *pos != c) {
		(*p)--;
		pos--;
	}

	return (pos >= str) ? pos : NULL;
}

#define MGD_FT_ELT_LEVEL 2

struct _ftstatus
{
	int entity;
	int pagefile;
	int level;
	int abort;
	int stopped;
	iconv_t cd;

	GString *name;
	GByteArray *value;

	mgd_ft_signal_handler signal_handler;
	mgd_ft_element_handler element_handler;
	mgd_ft_file_handler file_handler;
	mgd_ft_syntaxerror_handler syntaxerror_handler;
	void *client_data;
};

int mgd_ft_collect(char *startfile, struct _ftstatus *ftstatus);

int midgard_traverse_filetemplate(
      char *startfile, char *templatefile,
      mgd_ft_signal_handler signal_handler,
      mgd_ft_element_handler element_handler,
      mgd_ft_file_handler file_handler,
      mgd_ft_syntaxerror_handler
      syntaxerror_handler, void *client_data)
{
	static char path[MAXPATHLEN + 1];
	int len, status;
	char *stylestart;
	struct _ftstatus ftstatus;

	if (!startfile || *startfile != '/'
	    || strstr(startfile, "/../") != NULL) {
		return MGD_FT_ERR_PATH;
	}

	if (strlen(startfile) > MAXPATHLEN) {
		return MGD_FT_ERR_PATH;
	}

	strcpy(path, startfile);

	stylestart = NULL;
   if (templatefile != NULL && *templatefile != '\0') {
      len = strlen(startfile);
   } else if ((templatefile = strrchr(startfile, '.')) != NULL) {
		len = templatefile - startfile;
	}

	ftstatus.pagefile = 1;
	ftstatus.abort = 0;
	ftstatus.stopped = 0;
	ftstatus.name = g_string_new("");
	ftstatus.value = g_byte_array_new();
	ftstatus.signal_handler = signal_handler;
	ftstatus.element_handler = element_handler;
	ftstatus.file_handler = file_handler;
	ftstatus.syntaxerror_handler = syntaxerror_handler;
	ftstatus.client_data = client_data;

	do {
		status = mgd_ft_collect(path, &ftstatus);
		ftstatus.pagefile = 0;

		if (status != MGD_FT_ERR_NONE)
			return status;

		if (!ftstatus.stopped && templatefile
		    && (stylestart = mgd_strrchr(path, '/', &len)) != NULL) {
			strcpy(stylestart + 1, templatefile);
			len--;
		}
	} while (!ftstatus.stopped && stylestart != NULL);

	g_string_free(ftstatus.name, 1);
	g_byte_array_free(ftstatus.value, 1);

	return MGD_FT_ERR_NONE;
}

enum
{
	MGD_FT_ENTITY_PAGE,
	MGD_FT_ENTITY_ELT,
	MGD_FT_ENTITY_OTHER
};

void mgd_ft_elt_start(void *userdata, const char *name, const char **atts);
void mgd_ft_elt_end(void *userdata, const char *name);
void mgd_ft_cdata(void *userdata, const char *data, int len);

int mgd_ft_collect(char *startfile, struct _ftstatus *ftstatus)
{
	struct stat filestat;
	XML_Parser parser;
	FILE *fp;
	int done;
	char buf[BUFSIZ];
	int status;

	if (stat(startfile, &filestat) == -1) {
		if (ftstatus->pagefile || errno != ENOENT)
			return MGD_FT_ERR_ACCESS;
		return MGD_FT_ERR_NONE;
	}

	if (!S_ISREG(filestat.st_mode)) {
		return MGD_FT_ERR_ACCESS;
	}

	if (ftstatus->file_handler != NULL
	    && !ftstatus->file_handler(startfile, ftstatus->client_data))
		   return MGD_FT_ERR_ABORT;

	status = MGD_FT_ERR_NONE;

	ftstatus->entity = MGD_FT_ENTITY_OTHER;
	ftstatus->level = 0;
	ftstatus->abort = 0;
	g_string_assign(ftstatus->name, "");
	g_byte_array_set_size(ftstatus->value, 0);

	if ((fp = fopen(startfile, "r")) == NULL) {
		return ftstatus->pagefile ? MGD_FT_ERR_ACCESS : MGD_FT_ERR_NONE;
	}

	parser = XML_ParserCreate(NULL);
	XML_SetUserData(parser, ftstatus);
	XML_SetElementHandler(parser, mgd_ft_elt_start, mgd_ft_elt_end);
	XML_SetCharacterDataHandler(parser, mgd_ft_cdata);
	XML_SetUnknownEncodingHandler (parser, midgard_unknown_xml_encoding,
				 &ftstatus->cd);
				 
	ftstatus->cd = NULL;
	
	do {
		size_t len = fread(buf, 1, sizeof (buf), fp);
		done = len < sizeof (buf);

		if (!XML_Parse(parser, buf, len, done)) {
			status = MGD_FT_ERR_NWF;
			if (ftstatus->syntaxerror_handler)
				ftstatus->syntaxerror_handler
				   (XML_GetCurrentLineNumber(parser),
				    ftstatus->client_data);
			break;
		}

		if (ftstatus->abort)
			status = MGD_FT_ERR_ABORT;

	} while (!ftstatus->abort && !done);

	XML_ParserFree(parser);
	fclose(fp);
	
	if(ftstatus->cd != NULL) iconv_close(ftstatus->cd);
	return status;
}

const char *mgd_ft_get_att(char *name, const char **atts)
{
	if (atts == NULL)
		return NULL;

	while (*atts) {
		if (strcmp(name, *atts) == 0)
			return *(atts + 1);
		atts += 2;
	}

	return NULL;
}

void mgd_ft_elt_start(void *userdata, const char *eltname, const char **atts)
{
	struct _ftstatus *ftstatus = (struct _ftstatus *) userdata;
	const char *attname;

	if (ftstatus->abort)
		return;

	ftstatus->level++;

	if (ftstatus->level != MGD_FT_ELT_LEVEL)
		return;

	ftstatus->entity = MGD_FT_ENTITY_OTHER;

	if (strcmp(eltname, MGD_FT_ELTNAME_PAGE) == 0) {
		if (!ftstatus->pagefile && ftstatus->signal_handler) {
			ftstatus->abort =
			   !ftstatus->signal_handler
			   (MGD_FT_SIGNAL_PAGE_IN_STYLE, ftstatus->client_data);
		}
		if (!ftstatus->abort) {
			if (
			    (attname =
			     mgd_ft_get_att(MGD_FT_ELTNAME_PAGE_NAME,
					    atts)) != NULL && *attname != '\0') {
				ftstatus->entity = MGD_FT_ENTITY_PAGE;

				g_string_assign(ftstatus->name,
						attname != NULL ? attname : "");
				g_byte_array_set_size(ftstatus->value, 0);
			}
		}
	}
	else if (strcmp(eltname, MGD_FT_ELTNAME_ELT) == 0) {
		if ((attname = mgd_ft_get_att(MGD_FT_ELTNAME_ELT_NAME, atts)) !=
		    NULL && *attname != '\0') {
			ftstatus->entity = MGD_FT_ENTITY_ELT;

			g_string_assign(ftstatus->name,
					attname != NULL ? attname : "");
			g_byte_array_set_size(ftstatus->value, 0);
		}
	}
	else if (strcmp(eltname, MGD_FT_ELTNAME_AUTH) == 0) {
		if (ftstatus->signal_handler) ftstatus->abort =
		   !ftstatus->signal_handler(MGD_FT_SIGNAL_AUTH,
					     ftstatus->client_data);
	}
	else if (strcmp(eltname, MGD_FT_ELTNAME_STOP) == 0) {
		if (ftstatus->signal_handler) ftstatus->abort =
		   !ftstatus->signal_handler(MGD_FT_SIGNAL_STOP,
					     ftstatus->client_data);
		ftstatus->stopped = 1;
	}
	else if (strcmp(eltname, MGD_FT_ELTNAME_SHS) == 0) {
		if (ftstatus->signal_handler) ftstatus->abort =
		   !ftstatus->signal_handler(MGD_FT_SIGNAL_SKIP_HOST_STYLE,
					     ftstatus->client_data);
	}
}

void mgd_ft_elt_end(void *userdata, const char *name)
{
	struct _ftstatus *ftstatus = (struct _ftstatus *) userdata;
	char *fvalue;
	int len;

	if (ftstatus->abort)
		return;

	if (ftstatus->level == MGD_FT_ELT_LEVEL)
		switch (ftstatus->entity) {
			case MGD_FT_ENTITY_PAGE:
			if (ftstatus->value->len != 0) {
				if(ftstatus->cd != NULL)
				    fvalue = midgard_iconv(ftstatus->cd, 
					ftstatus->value->data, ftstatus->value->len, &len);
				else {
				    /* Append 0x0 only when iconv(3) isn't used */
				    g_byte_array_append(ftstatus->value, "", 1);
				    fvalue = ftstatus->value->data;
				}
				if (ftstatus->element_handler) {
					ftstatus->abort =
					   !ftstatus->element_handler
					   (MGD_FT_ELTNAME_PAGE_NAME,
					    ftstatus->name->str,
					    ftstatus->client_data);
					if (!ftstatus->abort)
						ftstatus->abort =
						   !ftstatus->element_handler
						   (MGD_FT_ELTNAME_PAGE_CONTENT,
						    fvalue,
						    ftstatus->client_data);
				}
				if(fvalue && (ftstatus->cd != NULL)) free(fvalue);
			}
			break;

			case MGD_FT_ENTITY_ELT:
			if (ftstatus->value->len != 0) {
				if(ftstatus->cd != NULL)
				    fvalue = midgard_iconv(ftstatus->cd, 
					ftstatus->value->data, ftstatus->value->len, &len);
				else {
				    /* Append 0x0 only when iconv(3) isn't used */
				    g_byte_array_append(ftstatus->value, "", 1);
				    fvalue = ftstatus->value->data;
				}
				if (ftstatus->element_handler)
					ftstatus->abort =
					   !ftstatus->element_handler(ftstatus->
								      name->str,
								      fvalue,
								      ftstatus->
								      client_data);
				if(fvalue && (ftstatus->cd != NULL)) free(fvalue);
			}
			break;

			default:
			break;
		}

	g_string_assign(ftstatus->name, "");
	g_byte_array_set_size(ftstatus->value, 0);

	ftstatus->level--;
}

void mgd_ft_cdata(void *userdata, const char *data, int len)
{
	struct _ftstatus *ftstatus = (struct _ftstatus *) userdata;

	if (ftstatus->abort)
		return;

	if (ftstatus->level == MGD_FT_ELT_LEVEL)
		switch (ftstatus->entity) {
			case MGD_FT_ENTITY_PAGE:
			case MGD_FT_ENTITY_ELT:
			g_byte_array_append(ftstatus->value, data, len);
			break;
		}
}

#endif /* HAVE_MIDGARD_FILETEMPLATES */
