/* $Id: repligard.c,v 1.10 2002/08/19 15:40:24 daniel Exp $
 *
 * midgard-lib: database access for Midgard clients
 *
 * Copyright (C) 2000 The Midgard Project ry
 * Copyright (C) 2000 Alexander Bokovoy and Aurora S.A.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "midgard/midgard.h"
#include "midgard/internal.h"
#include "defaults.h"
#include "midgard/parser.h"
#include "midgard/parsers.h"
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "md5.h"
#if MIDGARD_HAVE_UUID
#include <uuid/uuid.h>
#endif

#ifdef WIN32
#include <process.h>
#include "win32/win95nt.h"
#endif

extern char* repligard_magic_string; /* in midgard.c */

int mgd_create_repligard(midgard * mgd, const char *table,
			 const char *fields, const char *values, ...)
{
	int id;
	va_list args;
	va_start(args, values);
	id = mgd_vcreate_repligard(mgd, table, fields, values, args);
	va_end(args);
	return id;
}

int mgd_vcreate_repligard(midgard * mgd, const char *table,
			  const char *fields, const char *values, va_list args)
{
	const char *command;
	int rv;

	/* format command */
	command = mgd_format(mgd, mgd->tmp,
			     "INSERT INTO repligard (realm,$s) VALUES ($q,$s)",
			     fields, table, values);
	if (!command)
		return 0;

	/* execute command */
	rv = mgd_vexec(mgd, command, args);

	mgd_clear_pool(mgd->tmp);
	return rv ? mysql_insert_id(mgd->mysql) : 0;
}

int mgd_update_repligard(midgard * mgd, const char *table, int id,
			 const char *fields, ...)
{
	int rv;
	va_list args;
	va_start(args, fields);
	rv = mgd_vupdate_repligard(mgd, table, id, fields, args);
	va_end(args);
	return id;
}

int mgd_vupdate_repligard(midgard * mgd, const char *table, int id,
			  const char *fields, va_list args)
{
	const char *command;
	int rv;

	/* format command */
	command = mgd_format(mgd, mgd->tmp,
			     "UPDATE $s SET $s WHERE id=$d AND realm=$q",
			     "repligard", fields, id, table);
	if (!command)
		return 0;

	/* execute command */
	rv = mgd_vexec(mgd, command, args);

	mgd_clear_pool(mgd->tmp);
	return rv;
}

int mgd_delete_repligard(midgard * mgd, const char *table, int id)
{
	const char *command;
	int rv;

	/* format command */
	command =
	   mgd_format(mgd, mgd->tmp,
		      "DELETE FROM repligard WHERE realm=$q AND id=$d", table,
		      id);
	if (!command)
		return 0;

	/* execute command */
	rv = mgd_exec(mgd, command);

	mgd_clear_pool(mgd->tmp);
	return rv;
}

int mgd_delete_repligard_guid(midgard * mgd, const char *guid)
{
	const char *command;
	int rv;

	/* format command */
	command =
	   mgd_format(mgd, mgd->tmp,
		      "DELETE FROM repligard WHERE guid=$q", guid);
	if (!command)
		return 0;

	/* execute command */
	rv = mgd_exec(mgd, command);

	mgd_clear_pool(mgd->tmp);
	return rv;
}


char *mgd_repligard_guid(midgard * mgd, midgard_pool * pool, const char *table,
			int id)
{
	midgard_res *res;
	char *guid;
	res =
	   mgd_ungrouped_select(mgd, "guid", "repligard", "realm=$q AND id=$d",
				NULL, table, id);
	if (res && mgd_fetch(res)) {
		guid = mgd_strdup(pool, mgd_colvalue(res, 0));
		mgd_release(res);
		return guid;
	}
	return NULL;
}

char *mgd_repligard_changed(midgard * mgd, midgard_pool * pool, const char *table, int id)
{
	midgard_res *res;
	char* changed;
	res = mgd_ungrouped_select(mgd, "changed",
				   "repligard", "realm=$q AND id=$d",
				   NULL, table, id);
	if (res && mgd_fetch(res)) {
		changed = mgd_strdup(pool, mgd_colvalue(res, 0));
		mgd_release(res);
		return changed;
	}
	return NULL;
}

char *mgd_repligard_changed_guid(midgard * mgd, midgard_pool * pool, const char *guid)
{
	midgard_res *res;
	char* changed;
	res = mgd_ungrouped_select(mgd, "changed",
				   "repligard", "guid=$q",
				   NULL, guid);
	if (res && mgd_fetch(res)) {
		changed = mgd_strdup(pool, mgd_colvalue(res, 0));
		mgd_release(res);
		return changed;
	}
	return NULL;
}

char *mgd_repligard_updated(midgard * mgd, midgard_pool * pool, const char *table, int id)
{
	midgard_res *res;
	char* updated;
	res = mgd_ungrouped_select(mgd, "updated",
				   "repligard", "realm=$q AND id=$d",
				   NULL, table, id);
	if (res && mgd_fetch(res)) {
		updated = mgd_strdup(pool, mgd_colvalue(res, 0));
		mgd_release(res);
		return updated;
	}
	return NULL;
}

char *mgd_repligard_updated_guid(midgard * mgd, midgard_pool * pool, const char *guid)
{
	midgard_res *res;
	char* updated;
	res = mgd_ungrouped_select(mgd, "updated",
				   "repligard", "guid=$q",
				   NULL, guid);
	if (res && mgd_fetch(res)) {
		updated = mgd_strdup(pool, mgd_colvalue(res, 0));
		mgd_release(res);
		return updated;
	}
	return NULL;
}

char* mgd_repligard_table(midgard * mgd, midgard_pool* pool, const char *guid)
{
	midgard_res *res;
	char* table;
	res =
	   mgd_ungrouped_select(mgd, "realm", "repligard", "guid=$q",
				NULL, guid);
	if (res && mgd_fetch(res)) {
		table = mgd_strdup(pool, mgd_colvalue(res, 0));
		mgd_release(res);
		return table;
	}
	return NULL;
}

char* mgd_repligard_action(midgard * mgd, midgard_pool* pool, const char *guid)
{
	midgard_res *res;
	char* action;
	res =
	   mgd_ungrouped_select(mgd, "action", "repligard", "guid=$q",
				NULL, guid);
	if (res && mgd_fetch(res)) {
		action = mgd_strdup(pool, mgd_colvalue(res, 0));
		mgd_release(res);
		return action;
	}
	return NULL;
}

int mgd_repligard_id(midgard * mgd, const char *guid)
{
	midgard_res *res;
	int id;
	res =
	   mgd_ungrouped_select(mgd, "id", "repligard", "guid=$q",
				NULL, guid);
	if (res && mgd_fetch(res)) {
		id = mgd_sql2id(res, 0);
		mgd_release(res);
		return id;
	}
	return 0;
}

/*  Touches all records from table and set their timestamps in Repligard
    If record doesn't exist in Repligard, it will be created.
*/
void mgd_repligard_touch(midgard * mgd, const char *table)
{
	midgard_res *res;

	res = mgd_ungrouped_select(mgd, "id", table, NULL, NULL);
	if (res) {
		while (mgd_fetch(res)) {
			if (mgd_exists_bool(mgd, "repligard", "id=$d AND realm=$q",
				       mgd_sql2id(res, 0), table)) {
				UPDATE_REPLIGARD(mgd, table, mgd_sql2id(res, 0))
			}
			else {
				CREATE_REPLIGARD(mgd, table, mgd_sql2id(res, 0))
			}
		}
		mgd_release(res);
	}
}

static char md5tag[MIDGARD_REPLIGARD_TAG_LENGTH];

const char *mgd_create_guid(midgard *mgd, const char *table, int id)
#if MIDGARD_HAVE_UUID
{
uuid_t uu;
int     i;
unsigned char *cp;

   uuid_generate(uu);
   for(i=0, cp=(unsigned char*) &uu; i<16; i++) {
     snprintf(&md5tag[i*2], 3, "%02x", *cp++);
   }
   return (const char*)md5tag;
}
#else
{
    const char *str; 
    midgard_pool *pool;
    static int _blobcounter = 0;

    _blobcounter++; if (_blobcounter > 9999) { _blobcounter = 0; }

    pool = mgd_alloc_pool();
    str = mgd_format(mgd, pool, "re$spli$d$sga$drd$s.$d.$d",
		    table, (int)time(NULL), repligard_magic_string, id, 
		    mgd_version(), _blobcounter, (int)getpid());
    md5(str, md5tag);
    mgd_free_pool(pool);

    return (const char*)md5tag;
}
#endif
