#!/bin/sh
exec perl -x $0 $@

#!perl

use Getopt::Std;

$SIG{'__WARN__'} = sub { $wrongopts = 1; warn $_[0]; };
getopts('ch');
$SIG{'__WARN__'} = undef;
$wrongopts && exit(1);

$opt_c && $opt_h && die "Can't generate .h and .c\n";
!$opt_c && !$opt_h && die "Need something to do\n";

########### CONFIGURABLES ##########################
$unset = 'unset';

$commonprefix = "MIDGARD_OBJECT_";

%tablenames=(
	"article"          => "article",
	"blobs"            => "attachment",
	"element"          => "element",
	"event"            => "event",
	"eventmember"      => "eventmember",
	"file"             => "file",
	"grp"              => "group",
	"history"          => "history",
	"host"             => "host",
	"image"            => "image",
	"member"           => "member",
	"page"             => "page",
	"pageelement"      => "pageelement",
	"pagelink"         => "pagelink",
	"person"           => "person",
	"preference"       => "preference",
	"record_extension" => "parameter",
	"repligard"        => "repligard",
	"sitegroup"        => "sitegroup",
	"snippet"          => "snippet",
	"snippetdir"       => "snippetdir",
	"style"            => "style",
	"topic"            => "topic"
   );

########### END CONFIGURABLES ##########################

#### EEH: Add 'unset' #######
$tablenames{$unset} = '[' . $unset . ']';

@keys = sort keys %tablenames;
$count = $#keys + 1;

if ($opt_h) {
	print "/* Emile Heyns (http://www.midgard-project.org/): DO NOT EDIT! */\n";
	print "/* This is a generated files. Changes will simply be overwritten */\n";
	print "/* in the next update */\n";
   print "#ifndef MIDGARD_TABLENAMES_H\n";
   print "#define MIDGARD_TABLENAMES_H\n\n";

   foreach $int (@keys) {
      if (length($int) > $maxnamelen) { $maxnamelen = length($int); }
   }
   $maxnamelen += length($commonprefix);

   print sprintf("%-" . ($maxnamelen + 3) . "s   %d\n",
      "#define ${commonprefix}COUNT", $count);
   print "#define mgd_table_id_valid(id) (((id) >= 1) && ((id) <= $count))\n";
   print "typedef enum {\n";
   $i=0;
   $comma = ",";
   foreach $int (@keys) {
      $i++;
      if ($i == $count) { $comma = ""; }
      print sprintf("   %-${maxnamelen}s = %d$comma\n", $commonprefix . uc($int), $i);
   }
   print "} mgd_tableid_t;\n\n";

   print "extern const char *mgd_table_name["."];\n";
   print "extern const char *mgd_table_extname["."];\n";

   print "#endif\n";
}

if ($opt_c) {
	print "/* Emile Heyns (http://www.midgard-project.org/): DO NOT EDIT! */\n";
	print "/* This is a generated files. Changes will simply be overwritten */\n";
	print "/* in the next update */\n";
   print "#include \"midgard/midgard.h\"\n";

   print "const char *mgd_table_name["."] = {\n\t\tNULL";
   foreach $int (@keys) {
      $ext = $tablenames{$int};
      print ",\n\t\t\"$int\"";
   }
   print "\n};\n";

   print "const char *mgd_table_extname["."] = {\n\t\tNULL";
   foreach $int (@keys) {
      $ext = $tablenames{$int};
      print ",\n\t\t\"$ext\"";
   }
   print "\n};\n";
}
